use derive_more::{Add, AddAssign, Display, Sub, SubAssign};
use num_traits::float::FloatCore;
use crate::quantity::{DECI, Quantity, Value};

const KELVIN_CELSIUS_OFFSET: f64 = 273.15;

// ---------------------------------------------------------------------------------------------------------------------
// ----- Kelvins ------------------------
// ---------------------------------------------------------------------------------------------------------------------
#[derive(Copy, Clone, PartialEq, PartialOrd, Add, AddAssign, Sub, SubAssign, Display)]
#[display(fmt = "{} {}", _0, "Self::symbol()")]
#[repr(transparent)]
pub struct Kelvins<V: Value>(pub V);

impl <V: Value> Quantity<V> for Kelvins<V> {
    #[inline(always)]
    fn value(self) -> V {
        self.0
    }

    #[inline(always)]
    fn symbol() -> &'static str {
        "K"
    }
}

impl <V: Value> Kelvins<V> {
    #[inline]
    pub fn to_degrees_celsius(self) -> DegreesCelsius<V> {
        //TODO: Make const
        let offset = V::from_f64(KELVIN_CELSIUS_OFFSET).unwrap();
        DegreesCelsius(self.0 - offset)
    }
    
    #[inline]
    pub fn to_deci_kelvins(self) -> DeciKelvins<V> {
        let multiplier = V::from_u8(DECI).unwrap();
        DeciKelvins(self.0 * multiplier)
    }
}

// ---------------------------------------------------------------------------------------------------------------------
// ----- Deci Kelvins ------------------------
// ---------------------------------------------------------------------------------------------------------------------
#[derive(Copy, Clone, PartialEq, PartialOrd, Add, AddAssign, Sub, SubAssign, Display)]
#[display(fmt = "{} {}", _0, "Self::symbol()")]
#[repr(transparent)]
pub struct DeciKelvins<V: Value>(pub V);

impl <V: Value> Quantity<V> for DeciKelvins<V> {
    #[inline(always)]
    fn value(self) -> V {
        self.0
    }

    #[inline(always)]
    fn symbol() -> &'static str {
        "dK"
    }
}

impl<V: Value> DeciKelvins<V> {
    #[inline]
    pub fn to_kelvins(self) -> Kelvins<V> {
        let divisor = V::from_u8(DECI).unwrap();
        Kelvins(self.0 / divisor)
    }
}

// ---------------------------------------------------------------------------------------------------------------------
// ----- Degrees Celsius ------------------------
// ---------------------------------------------------------------------------------------------------------------------
#[derive(Copy, Clone, PartialEq, PartialOrd, Add, AddAssign, Sub, SubAssign, Display)]
#[display(fmt = "{} {}", _0, "Self::symbol()")]
#[repr(transparent)]
pub struct DegreesCelsius<V: Value>(pub V);

impl <V: Value> Quantity<V> for DegreesCelsius<V> {
    #[inline(always)]
    fn value(self) -> V {
        self.0
    }

    #[inline(always)]
    fn symbol() -> &'static str {
        "℃"
    }
}

impl <V: Value> DegreesCelsius<V> {
    #[inline]
    pub fn to_kelvins(self) -> Kelvins<V> {
        //TODO: Make const
        let offset = V::from_f64(KELVIN_CELSIUS_OFFSET).unwrap();
        Kelvins(self.0 + offset)
    }
    
    #[inline]
    pub fn to_deci_degrees_celsius(self) -> DeciDegreesCelsius<V> {
        let multiplier = V::from_u8(DECI).unwrap();
        DeciDegreesCelsius(self.0 * multiplier)
    }
}

// ---------------------------------------------------------------------------------------------------------------------
// ----- Deci Degrees Celsius ------------------------
// ---------------------------------------------------------------------------------------------------------------------
#[derive(Copy, Clone, PartialEq, PartialOrd, Add, AddAssign, Sub, SubAssign, Display)]
#[display(fmt = "{} {}", _0, "Self::symbol()")]
#[repr(transparent)]
pub struct DeciDegreesCelsius<V: Value>(pub V);

impl <V: Value> Quantity<V> for DeciDegreesCelsius<V> {
    #[inline(always)]
    fn value(self) -> V {
        self.0
    }

    #[inline(always)]
    fn symbol() -> &'static str {
        "d℃"
    }
}

impl<V: Value> DeciDegreesCelsius<V> {
    #[inline]
    pub fn to_degrees_celsius(self) -> DegreesCelsius<V> {
        let divisor = V::from_u8(DECI).unwrap();
        DegreesCelsius(self.0 / divisor)
    }
}

// ---------------------------------------------------------------------------------------------------------------------
// ----- Tests ------------------------
// ---------------------------------------------------------------------------------------------------------------------
#[cfg(test)]
mod tests {
    use float_cmp::assert_approx_eq;
    use super::*;

    #[test]
    fn convert_f32() {
        let kelvins: Kelvins<f32> =  298.15.kelvins();
        let deci_kelvins: DeciKelvins<f32> = 2_981.5.deci_kelvins();
        let celsius: DegreesCelsius<f32> = 25.0.degrees_celsius();
        let deci_celsius: DeciDegreesCelsius<f32> = 250.0.deci_degrees_celsius();
        
        assert_approx_eq!(f32, kelvins.to_degrees_celsius().0, celsius.0);
        assert_approx_eq!(f32, celsius.to_kelvins().0, kelvins.0);

        assert_approx_eq!(f32, deci_kelvins.to_kelvins().0, kelvins.0);
        assert_approx_eq!(f32, kelvins.to_deci_kelvins().0, deci_kelvins.0);
        
        assert_approx_eq!(f32, deci_celsius.to_degrees_celsius().0, celsius.0);
        assert_approx_eq!(f32, celsius.to_deci_degrees_celsius().0, deci_celsius.0);
    }
    
    #[test]
    fn convert_u16() {
        let kelvins: Kelvins<u16> = 298.kelvins();
        let degrees_celsius: DegreesCelsius<u16> = 25.degrees_celsius();
        
        assert_eq!(degrees_celsius.0, kelvins.to_degrees_celsius().0);
    }
}
